
# Functions for the true model
# Mean response surface
MU.Y <- function(par, X){
  b0 <- par[1]; b1 <- par[2]; b2 <- par[3]
  # Max is -b/2a
  b2*X^2 + b1*X + b0
}

# Measurement noise
SIG.Y <- function(par){par[4]}

# Mean nominal-to-actual curve
MU.X <- function(par, Z){
  par[5]*Z
}

# nominal-to-actual noise
SIG.X <- function(par, Z){
  par[6]*sqrt(Z)
}


# Function for generating data from true model for specific 
# noise level (tau), a chosen number of observations (num_obs),
# and a chosen number of surrogate observations (num_surr).
Gen_data <- function(par, num_obs=20, num_surr=50, Z=NULL, lb, ub){
  
  if (is.null(Z)){
  zz <- seq(lb, ub, length.out=num_obs+2)[2:(num_obs+1)]
  }else{
    zz <- Z
    num_obs <- length(zz)
  }
  
  xx <- rnorm(num_obs, mean=MU.X(par, zz), sd=SIG.X(par, zz))
  
  yy <- rnorm(num_obs, mean=MU.Y(par, xx), sd=SIG.Y(par))
  
  zz_surr <- seq(lb, ub, length.out=num_surr+2)[2:(num_surr+1)]
  xx_surr <- rnorm(num_surr, mean=MU.X(par, zz_surr), 
                   sd=SIG.X(par, zz_surr))
  
  return(list(Z=zz, X=xx, Y=yy, Zplus=zz_surr, Xplus=xx_surr))  
}

## Generate data for an estimated optimal z basd on true parameters
Gen_post <- function(z_opts, par, num_sim=1e6){
  z1 <- z_opts[1]; z2 <- z_opts[2]
  
  xx1 <- rnorm(num_sim, mean=MU.X(par, z1), sd=SIG.X(par, z1))
  yy1 <- rnorm(num_sim, mean=MU.Y(par, xx1), sd=SIG.Y(par))
  
  xx2 <- rnorm(num_sim, mean=MU.X(par, z2), sd=SIG.X(par, z2))
  yy2 <- rnorm(num_sim, mean=MU.Y(par, xx2), sd=SIG.Y(par))
  
  return(list(Y1=yy1, Y2=yy2))  
}


#### Functions for the coefficients of each model
Models <- function(dataset){
  ## The dataset for full model
  dataset_1 <- list(D = as.matrix(cbind(rep(1, length(dataset$X)), dataset$X, dataset$X^2)),
                    Y = matrix(dataset$Y, ncol=1),
                    X = c(dataset$X, dataset$Xplus),
                    Z = c(dataset$Z, dataset$Zplus))
  
  ## The dataset for zero-internal-noise model
  dataset_2 <- list(D = as.matrix(cbind(rep(1, length(dataset$Z)), dataset$Z, dataset$Z^2)),
                    Y = matrix(dataset$Y, ncol=1))
  
  
  #### Log-likelihood, prior and posterior functions
  ## Log-likelihood function for regression parameters
  llike_reg <- function(params, data){
    nn <- length(params)
    
    bet <- matrix(params[-nn], ncol=1)
    sig <- exp(params[nn])
    
    D <- data$D
    Y <- data$Y
    EE <- as.numeric(Y - D%*%bet)
    
    ll <- sum(dnorm(EE, mean=0, sd=sig, log=TRUE))
    return(ll)
  }
  
  ## Log-prior function for regression parameters
  lprior_reg <- function(params){
    lpr <- sum(dnorm(params, mean=0, sd=1e5, log=TRUE))
    return(lpr)
  }
  
  ## Log-likelihood function for hyperparameters on the log-scale
  llike_hyper <- function(params, data){
    gam <- exp(params[1])
    tau <- exp(params[2])
    
    X <- data$X
    Z <- data$Z
    
    ll <- sum(dnorm(X, mean=gam*Z, sd=tau*sqrt(Z), log=TRUE))
    return(ll)
  }
  
  ## Log-prior function for hyperparameters on the log-scale
  lprior_hyper <- function(params){
    lpr <- sum(dnorm(params, mean=0, sd=1e5, log=TRUE))
    return(lpr)
  }
  
  ## Log-posterior for regression parameters
  lpost_reg <- function(params, data){
    lpst <- llike_reg(params, data) + lprior_reg(params)
    
    return(lpst)
  }
  
  ## Log-posterior for hyper-parameters
  lpost_hyper <- function(params, data){
    lpst <- llike_hyper(params, data) + lprior_hyper(params)
    
    return(lpst)
  }
  
  #browser()
  #### Parameter estimation for full model
  ## Parameter initialization
  num_reg_pars <- 1 + ncol(dataset_1$D)
  num_hyper_pars <- 2
  
  init_reg_pars <- runif(num_reg_pars, -1, 1)
  init_hyper_pars <- runif(num_hyper_pars, -1, 1)
  
  ## Finding the parameter MAPs
  fit_reg <- optim(init_reg_pars, lpost_reg, data=dataset_1, method="BFGS",
                   control=list(fnscale=-1, maxit=1e5))
  fit_hyper <- optim(init_hyper_pars, lpost_hyper, data=dataset_1, method="BFGS",
                     control=list(fnscale=-1, maxit=1e5))
  
  est_bets <- fit_reg$par[-num_reg_pars]
  est_sig <- exp(fit_reg$par[num_reg_pars])
  
  est_gam <- exp(fit_hyper$par[num_hyper_pars-1])
  est_tau <- exp(fit_hyper$par[num_hyper_pars])
  
  ## Saving
  coeff_1 <- c(est_bets, est_sig, est_gam, est_tau)
  
  
  #### Parameter estimation for zero-internal-noise model
  ## Parameter initialization
  num_reg_pars_2 <- 1 + ncol(dataset_2$D)
  init_reg_par_2 <- runif(num_reg_pars_2, -1, 1)
  
  ## Finding the parameter MAPs
  fit_reg_2 <- optim(init_reg_par_2, lpost_reg, data=dataset_2, method="BFGS",
                     control=list(fnscale=-1, maxit=1e5))
  
  est_bets_2 <- fit_reg_2$par[c(1:(num_reg_pars_2-1))]
  est_sig_2 <- exp(fit_reg_2$par[num_reg_pars_2])
  
  ## Saving
  coeff_2 <- c(est_bets_2, est_sig_2)
  
  return(list(mod1 = coeff_1, mod2 = coeff_2))
}


SAN1 <- function(par, zz){
  b0 <- par[1]; b1 <- par[2]; b2 <- par[3]; sig <- par[4]
  gg <- par[5]; tau <- par[6]
  
  signal <- (b1*gg + b2*(tau^2))*zz + b2*(gg^2)*(zz^2) + b0
  varr <- sig^2 + ((b1*tau)^2)*zz + 
    (4*b1*b2*gg*(tau^2) + 2*(b2^2)*(tau^4))*(zz^2) +
    4*((b2*gg*tau)^2)*(zz^3)
  
  return(list(signal=signal, noise=sqrt(varr)))
}

SAN2 <- function(par, zz){
  c0 <- par[1]; c1 <- par[2]; c2 <- par[3]; sig_c <- par[4]
  
  signal <- c1*zz + c2*(zz^2) + c0
  
  return(list(signal=signal, noise=rep(sig_c, length(zz))))
}

# Reciprocal coefficient of variation
SNR1 <- function(par, zz){
  san <- SAN1(par, zz)
  
  return(san$signal/san$noise)
}

SNR2 <- function(par, zz){
  san <- SAN2(par, zz)
  
  return(san$signal/san$noise)
}


MVK1 <- function(par, zz){
  b0 <- par[1]; b1 <- par[2]; b2 <- par[3]; sig <- par[4]
  gg <- par[5]; tau <- par[6]
  
  kk <- 1
  mz <- (b1/(2*b2) + gg*zz)^2
  
  mu_x <- b2*(kk*(tau^2)*zz + mz) + (b0 - (b1^2)/(4*b2))
  #mu_x <- b0 + b1*gg*zz + b2*((gg^2)*(zz^2) + (tau^2)*zz)
  var_x <- 2*(b2^2)*(tau^2)*zz*(kk*(tau^2)*zz + 2*mz)
  skew_x <- 8*(b2^3)*(tau^4)*(zz^2)*(kk*(tau^2)*zz + 3*mz)
  kurt_x <- 12*(b2^4)*(tau^4)*(zz^2)*(kk*(tau^2)*zz + 2*mz)^2 +
    48*(b2^4)*(tau^6)*(zz^3)*(kk*(tau^2)*zz + 4*mz)
  
  return(list(m=mu_x, v=var_x, s=skew_x, k=kurt_x))
}

MVK2 <- function(par, zz){
  c0 <- par[1]; c1 <- par[2]; c2 <- par[3]; sig_c <- par[4]
  
  mu_x <- c0 + c1*zz + c2*(zz^2)
  var_x <- rep(0, length(zz))
  skew_x <- rep(0, length(zz))
  kurt_x <- rep(0, length(zz))
  
  return(list(m=mu_x, v=var_x, s=skew_x, k=kurt_x))
}


PDM1 <- function(pars, zz){
  par <- pars[[1]]
  b0 <- par[1]; b1 <- par[2]; b2 <- par[3]; sig <- par[4]
  gg <- par[5]; tau <- par[6]
  
  z_star <- -(b1*gg + b2*(tau^2))/(2*b2*(gg^2)) # -b1/(2*b2)
  mu_star <- MVK1(par, z_star)$m
  
  mvk <- MVK1(par, zz)
  s_star <- (mu_star - mvk$m)/sig
  
  f_0 <- (1 - pnorm(s_star))
  f_2 <- s_star*dnorm(s_star)/(sig^2)
  f_3 <- (s_star^2 - 1)*dnorm(s_star)/(sig^3)
  f_4 <- (s_star^2 - 3)*s_star*dnorm(s_star)/(sig^4)
  
  pdm_1 <- f_0
  pdm_2 <- mvk$v*f_2/2
  pdm_3 <- mvk$s*f_3/6
  pdm_4 <- mvk$k*f_4/24
  
  pdm <- pdm_1 + pdm_2 + pdm_3 #+ pdm_4
  
  return(pdm)
}

PDM2 <- function(pars, zz){
  par <- pars[[2]]
  c0 <- par[1]; c1 <- par[2]; c2 <- par[3]; sig_c <- par[4]
  
  z_star <- -c1/(2*c2) # -c1/(2*c2)
  mu_star <- MVK2(par, z_star)$m
  
  mvk <- MVK2(par, zz)
  s_star <- (mu_star - mvk$m)/sig_c
  
  f_0 <- (1 - pnorm(s_star))
  f_2 <- s_star*dnorm(s_star)/(sig_c^2)
  f_3 <- (s_star^2 - 1)*dnorm(s_star)/(sig_c^3)
  f_4 <- (s_star^2 - 3)*s_star*dnorm(s_star)/(sig_c^4)
  
  pdm_1 <- f_0
  pdm_2 <- mvk$v*f_2/2
  pdm_3 <- mvk$s*f_3/6
  pdm_4 <- mvk$k*f_4/24
  
  pdm <- pdm_1 + pdm_2 + pdm_3 #+ pdm_4
  
  return(pdm)
}

EDM1 <- function(pars, zz){
  #browser()
  par <- pars[[1]]
  b0 <- par[1]; b1 <- par[2]; b2 <- par[3]; sig <- par[4]
  gg <- par[5]; tau <- par[6]
  
  z_star <- -(b1*gg + b2*(tau^2))/(2*b2*(gg^2)) # -b1/(2*b2)
  mu_star <- MVK1(par, z_star)$m
  
  mvk <- MVK1(par, zz)
  s_star <- (mu_star - mvk$m)/sig
  
  
  edm_1 <- mvk$m*(1 - pnorm(s_star)) + sig*dnorm(s_star)
  edm_2 <- (mvk$v/2)*(1 + (mu_star/sig)*s_star)*dnorm(s_star)/sig
  edm_3 <- (mvk$s/6)*((mu_star/sig)*(s_star^2) -
                        (mvk$m/sig))*dnorm(s_star)/(sig^2)
  
  edm <- edm_1 + edm_2 + edm_3
  
  return(edm)
}

EDM2 <- function(pars, zz){
  par <- pars[[2]]
  c0 <- par[1]; c1 <- par[2]; c2 <- par[3]; sig_c <- par[4]
  
  z_star <- -c1/(2*c2) # -c1/(2*c2)
  mu_star <- MVK2(par, z_star)$m
  
  mvk <- MVK2(par, zz)
  s_star <- (mu_star - mvk$m)/sig_c
  
  edm_1 <- mvk$m*(1 - pnorm(s_star)) + sig_c*dnorm(s_star)
  edm_2 <- (mvk$v/2)*(1 + (mu_star/sig_c)*s_star)*dnorm(s_star)/sig_c
  edm_3 <- (mvk$s/6)*((mu_star/sig_c)*(s_star^2) -
                        (mvk$m/sig_c))*dnorm(s_star)/(sig_c^2)
  
  edm <- edm_1 + edm_2 + edm_3
  
  return(edm)
}


PDX1 <- function(pars, zz){
  #browser()
  par <- pars[[1]]
  b0 <- par[1]; b1 <- par[2]; b2 <- par[3]; sig <- par[4]
  gg <- par[5]; tau <- par[6]
  
  z_star <- -(b1*gg + b2*(tau^2))/(2*b2*(gg^2)) # -b1/(2*b2)
  mu_star <- MVK1(par, z_star)$m
  
  mu_x <- b0 + b1*gg*zz + b2*(gg^2)*(zz^2)
  s_star <- (mu_star - mu_x)/sig
  
  f_0 <- (1 - pnorm(s_star))
  f_2 <- dnorm(s_star)*(2*b2/sig + (((b1 + 2*b2*gg*zz)/sig)^2)*s_star)
  
  pdm_1 <- f_0
  pdm_2 <- (tau^2)*zz*f_2/2
  
  pdm <- pdm_1 + pdm_2
  
  return(pdm)
}

PDX2 <- function(pars, zz){
  par <- pars[[2]]
  c0 <- par[1]; c1 <- par[2]; c2 <- par[3]; sig_c <- par[4]
  
  z_star <- -c1/(2*c2) # -c1/(2*c2)
  mu_star <- MVK2(par, z_star)$m
  
  mu_x <- c0 + c1*zz + c2*(zz^2)
  s_star <- (mu_star - mu_x)/sig_c
  
  f_0 <- (1 - pnorm(s_star))
  
  pdm_1 <- f_0
  pdm_2 <- 0
  
  pdm <- pdm_1 + pdm_2
  
  return(pdm)
}

XREF <- function(pars, NN, lb, ub){
  par <- pars[[1]]
  gg <- par[5]; tau <- par[6]
  
  ZZ <- runif(NN, min=lb, max=ub)
  XX <- rnorm(NN, mean=gg*ZZ, sd=tau*sqrt(ZZ))
  
  HH <- sapply(1:NN, function(ii){
    dnorm(XX, mean=gg*ZZ[ii], sd=tau*sqrt(ZZ[ii]))       
  }) 
  
  return(list(XX=XX, ZZ=ZZ, HH=matrix(HH, ncol=NN)))
}
  
PDMC1 <- function(pars, xhRef, zz){
  #browser()
  set.seed(1234)
  
  par <- pars[[1]]
  b0 <- par[1]; b1 <- par[2]; b2 <- par[3]; sig <- par[4]
  gg <- par[5]; tau <- par[6]
  
  xRef <- xhRef$XX
  zRef <- xhRef$ZZ
  hRef <- xhRef$HH
  NN <- length(zRef)
  
  allZ <- c(zRef, zz)  
  extraX <- rnorm(length(zz), mean=gg*zz, sd=tau*sqrt(zz))
  allX <- c(xRef, extraX)
  
  HH_right <- sapply(1:length(zz), function(ii){
    dnorm(xRef, mean=gg*zz[ii], sd=tau*sqrt(zz[ii]))
  }) 
  HH_right <- matrix(HH_right, ncol=length(zz))
  
  HH_below <- t(sapply(extraX, function(x){
    dnorm(rep(x, NN+length(zz)), mean=gg*allZ, sd=tau*sqrt(allZ))
  }))
    
  newH <- allH <- rbind(cbind(hRef, HH_right), HH_below)

  
  inf.idx <- which(is.infinite(allH), arr.ind=TRUE, useNames=FALSE)  
  if (nrow(inf.idx) >0 ){
    for (kk in 1:nrow(inf.idx)){
      newH[inf.idx[kk, 1], inf.idx[kk, 2]] <- 1e20
    }
  }
  
  #   z_star <- -(b1*gg + b2*(tau^2))/(2*b2*(gg^2)) # -b1/(2*b2)
  #   mu_star <- MVK1(par, z_star)$m
  x_star <- -b1/(2*b2)
  mu_star <- b0 + b1*x_star + b2*(x_star^2)
  
  mu_Ref <- b0 + b1*allX + b2*(allX^2)
  s_star <- (mu_star - mu_Ref)/sig
  
  #   ff <- sapply(zz, function(z){
  #     lf <- dnorm(xRef, mean=gg*z, sd=tau*sqrt(z), log=TRUE)
  #     lf_max <- max(lf)
  #     
  #     if(lf_max==Inf){
  #       lf[lf==lf_max] <- 0
  #     }else if(lf_max==-Inf){
  #       lf <- 0*xRef
  #     }else{
  #       lf <- lf - lf_max
  #     }
  #     
  #     f0 <- exp(lf)
  #     
  #     return(f0/sum(f0))
  #   })
  
  ff <- sapply(1:length(zz), function(jj){
    lf <- as.numeric(newH[, (NN+jj)])/rowSums(newH)
    
    return(lf)
  })
  
  FF <- matrix(ff, ncol=length(zz))
  FF <- t(t(FF)/colSums(FF))
  pdm <- (1 - pnorm(s_star))
  pdmc <- as.numeric(pdm %*% FF)
  
  #browser()
  return(pdmc)
}


EDMC1 <- function(pars, xhRef, zz){
  #browser()
  set.seed(1234)
  
  par <- pars[[1]]
  b0 <- par[1]; b1 <- par[2]; b2 <- par[3]; sig <- par[4]
  gg <- par[5]; tau <- par[6]
  
  xRef <- xhRef$XX
  zRef <- xhRef$ZZ
  hRef <- xhRef$HH
  NN <- length(zRef)
  
  allZ <- c(zRef, zz)  
  extraX <- rnorm(length(zz), mean=gg*zz, sd=tau*sqrt(zz))
  allX <- c(xRef, extraX)
  
  HH_right <- sapply(1:length(zz), function(ii){
    dnorm(xRef, mean=gg*zz[ii], sd=tau*sqrt(zz[ii]))
  }) 
  HH_right <- matrix(HH_right, ncol=length(zz))
  
  HH_below <- t(sapply(extraX, function(x){
    dnorm(rep(x, NN+length(zz)), mean=gg*allZ, sd=tau*sqrt(allZ))
  }))
  
  newH <- allH <- rbind(cbind(hRef, HH_right), HH_below)
  
  
  inf.idx <- which(is.infinite(allH), arr.ind=TRUE, useNames=FALSE)
  #browser()
  if (nrow(inf.idx) >0 ){
  for (kk in 1:nrow(inf.idx)){
    newH[inf.idx[kk, 1], inf.idx[kk, 2]] <- 1e20
  }
  }
  #   z_star <- -(b1*gg + b2*(tau^2))/(2*b2*(gg^2)) # -b1/(2*b2)
  #   mu_star <- MVK1(par, z_star)$m
  x_star <- -b1/(2*b2)
  mu_star <- b0 + b1*x_star + b2*(x_star^2)
  
  mu_Ref <- b0 + b1*allX + b2*(allX^2)
  s_star <- (mu_star - mu_Ref)/sig
  
  #   ff <- sapply(zz, function(z){
  #     lf <- dnorm(xRef, mean=gg*z, sd=tau*sqrt(z), log=TRUE)
  #     lf_max <- max(lf)
  #     
  #     if(lf_max==Inf){
  #       lf[lf==lf_max] <- 0
  #     }else if(lf_max==-Inf){
  #       lf <- 0*xRef
  #     }else{
  #       lf <- lf - lf_max
  #     }
  #     
  #     f0 <- exp(lf)
  #     
  #     return(f0/sum(f0))
  #   })
  
  ff <- sapply(1:length(zz), function(jj){
    lf <- as.numeric(newH[, (NN+jj)])/rowSums(newH)
    
    return(lf)
  })
  
  FF <- matrix(ff, ncol=length(zz))
  FF <- t(t(FF)/colSums(FF))
  edm <- mu_Ref*(1 - pnorm(s_star)) + sig*dnorm(s_star)
  edmc <- as.numeric(edm %*% FF)
  
  return(edmc)
}


CVR <- function(pars, lb, ub){
  gg1 <- pars[5]; tt1 <- pars[6]
  res <- 1/sqrt(1 + 6*((tt1/gg1)^2)*(lb+ub)/((ub-lb)^2))
  
  return(res)
}

SFN1 <- function(pars, lb, ub){
  gg1 <- pars[5]; tt1 <- pars[6]
  res <- 1 - 1/sqrt(1 + 6*((tt1/gg1)^2)*(lb+ub)/((ub-lb)^2))
  
  return(res)
}

SFN2 <- function(pars, lb, ub){
  gg1 <- pars[5]; tt1 <- pars[6]
  res <- 6*(tt1^2)*(lb+ub)/((gg1*(ub-lb))^2 + 6*(tt1^2)*(lb+ub))
  
  return(res)
}

ISFN1 <- function(ss, gam, lb, ub){
  tt <- sqrt((2*ss - ss^2)/(6*(lb+ub)))*gam*(ub-lb)*(1/(1-ss))
  
  return(tt)
}

ISFN2 <- function(ss, gam, lb, ub){
  tt <- sqrt((1/(6*(lb+ub)))*(ss/(1-ss)))*gam*(ub-lb)
  
  return(tt)
}

######### End of functions
