
#### Functions for replicating Remillard et al (2016)
## Function for re-arranging model terms
name_rarr <- function(name, split=':'){
  name_vec <- strsplit(name, split=split)
  sapply(1:length(name), function(i){
    name_ord <- order(substring(name_vec[[i]],1,1))
    return(paste0(name_vec[[i]][name_ord], collapse=split))
  })
}

## Function for selecting best model dataframe based on criterion type: 'bic', 'cp' or 'adjr2'
best_terms <- function(model, type, rank){
  Mdat0 <- as.data.frame(summary(model)$which, 
                         row.names=1:nrow(summary(model)$which))
  names(Mdat0) <- name_rarr(names(Mdat0))
  
  if(type=='bic'){
    yy <- summary(model)$bic
  }else if(type=='cp'){
    yy <- summary(model)$cp
  }else if(type=='adjr2'){
    yy <- summary(model)$adjr2
  }else{
    print("Error: only choose from type=('bic', 'cp','adjr2')")
    break
  }
  
  zz <- sort(yy)[rank]
  
  best_names <- names(Mdat0)[as.logical(Mdat0[which(yy == zz),])]
  return(best_names)
}

## Function for finding the row of the best fit (based on criterion type) 
## within the best subset regression plots
model_choice <- function(model, type='bic', rank){
  # The full model logical matrix
  Mdat <- as.data.frame(summary(model)$which, 
                        row.names=1:nrow(summary(model)$which))
  names(Mdat) <- name_rarr(names(Mdat))
  
  # The best fit terms
  bestTerms <- best_terms(model=model, type=type, rank=rank)
  
  # Create a logical vector for the best fit terms using
  # the terms of the full model
  col.idx <- names(Mdat) %in% bestTerms
  # Which rows match the best fit logical vector
  row.idx <- sapply(1:nrow(Mdat), function(rr){
    all(Mdat[rr,] == col.idx)
  })
  row.num <- which(row.idx)
  
  # The BIC
  best_BIC <- summary(model)$bic[row.num]
  vline_BIC <- which(sort(summary(model)$bic, 
                          decreasing=T)==best_BIC)
  
  # Mallow's Cp
  best_Cp <- summary(model)$cp[row.num]
  vline_Cp <- which(sort(summary(model)$cp, 
                         decreasing=T)==best_Cp)
  
  # Adjusted R2
  best_adjr2 <- summary(model)$adjr2[row.num]
  vline_adjr2 <- which(sort(summary(model)$adjr2)==best_adjr2)
  
  return(list(bic=vline_BIC, cp=vline_Cp, adjr2=vline_adjr2, bestTerms = bestTerms))
}



#### Functions for the real noisy data
## Function for generating reference X's (noisy input) for Monte-Carlo optimization
XREF <- function(pars, NN, lb, ub){
  alp <- pars[1]; gam <- pars[2]
  
  ZZ <- runif(NN, min=lb, max=ub)
  XX <- rgamma(NN, shape=alp*ZZ, rate=rep(gam, NN))
  
  HH <- sapply(1:NN, function(ii){
    dgamma(XX, shape=alp*ZZ[ii], rate=gam)       
  }) 
  
  return(list(XX=XX, ZZ=ZZ, HH=matrix(HH, ncol=NN)))
}


## Objective function: Approximate Tail Probability
PDMC1 <- function(pars, xhRef, dom, y.star){
  set.seed(1234)
  
  alp <- pars[1]; gam <- pars[2]; sig <- pars[3]
  b0 <- pars[4]; b1 <- pars[5]; b2 <- pars[6]
  b3 <- pars[7]; b4 <- pars[8]; b5 <- pars[9]
  b6 <- pars[10]; b7 <- pars[11]
  
  zz <- dom[1]; cc <- dom[2]; vo <- dom[3]
  mm <- cc*vo
  
  
  xRef <- xhRef$XX
  zRef <- xhRef$ZZ
  hRef <- xhRef$HH
  NN <- length(zRef)
  
  allZ <- c(zRef, zz)  
  extraX <- rgamma(1, shape=alp*zz, rate=gam)
  allX <- c(xRef, extraX)
  
  HH_right <- dgamma(xRef, shape=alp*zz, rate=gam)
  HH_right <- matrix(HH_right, ncol=1)
  
  HH_below <-  dgamma(rep(extraX, NN+1), shape=alp*allZ, rate=rep(gam, NN+1))
  
  newH <- allH <- rbind(cbind(hRef, HH_right), HH_below)
  
  
  inf.idx <- which(is.infinite(allH), arr.ind=TRUE, useNames=FALSE)
  
  if (nrow(inf.idx) >0 ){
    for (kk in 1:nrow(inf.idx)){
      newH[inf.idx[kk, 1], inf.idx[kk, 2]] <- 1e20
    }
  }
  
  mu_star <- y.star
  TB <- 0; TC <- 0   ## Sample A only
  mu_Ref <- b0*cc^2 + (b1 + b2*TB + b3*TC + b4*cc)*allX + (b5 + b6*TC + b7*mm)*(allX^2)
  
  s_star <- (mu_star - mu_Ref)/sig
  
  ff <- sapply(1:length(zz), function(jj){
    lf <- as.numeric(newH[, (NN+jj)])/rowSums(newH)
    
    return(lf)
  })
  
  FF <- matrix(ff, ncol=length(zz))
  FF <- t(t(FF)/colSums(FF))
  pdm <- (1 - pnorm(s_star))
  pdmc <- as.numeric(pdm %*% FF)
  
  return(pdmc)
}

## Objective function: Approximate Tail Expectation
EDMC1 <- function(pars, xhRef, dom, y.star){
  set.seed(1234)
  
  alp <- pars[1]; gam <- pars[2]; sig <- pars[3]
  b0 <- pars[4]; b1 <- pars[5]; b2 <- pars[6]
  b3 <- pars[7]; b4 <- pars[8]; b5 <- pars[9]
  b6 <- pars[10]; b7 <- pars[11]
  
  zz <- dom[1]; cc <- dom[2]; vo <- dom[3]
  mm <- cc*vo
  
  
  xRef <- xhRef$XX
  zRef <- xhRef$ZZ
  hRef <- xhRef$HH
  NN <- length(zRef)
  
  allZ <- c(zRef, zz)  
  extraX <- rgamma(1, shape=alp*zz, rate=gam)
  allX <- c(xRef, extraX)
  
  HH_right <- dgamma(xRef, shape=alp*zz, rate=gam)
  HH_right <- matrix(HH_right, ncol=1)
  
  HH_below <-  dgamma(rep(extraX, NN+1), shape=alp*allZ, rate=rep(gam, NN+1))
  
  newH <- allH <- rbind(cbind(hRef, HH_right), HH_below)
  
  
  inf.idx <- which(is.infinite(allH), arr.ind=TRUE, useNames=FALSE)
  
  if (nrow(inf.idx) >0 ){
    for (kk in 1:nrow(inf.idx)){
      newH[inf.idx[kk, 1], inf.idx[kk, 2]] <- 1e20
    }
  }
  
  mu_star <- y.star
  TB <- 0; TC <- 0   ## Sample A only
  mu_Ref <- b0*cc^2 + (b1 + b2*TB + b3*TC + b4*cc)*allX + (b5 + b6*TC + b7*mm)*(allX^2)
  
  s_star <- (mu_star - mu_Ref)/sig
  
  ff <- sapply(1:length(zz), function(jj){
    lf <- as.numeric(newH[, (NN+jj)])/rowSums(newH)
    
    return(lf)
  })
  
  FF <- matrix(ff, ncol=length(zz))
  FF <- t(t(FF)/colSums(FF))
  edm <- mu_Ref*(1 - pnorm(s_star)) + sig*dnorm(s_star)
  edmc <- as.numeric(edm %*% FF)
  
  return(edmc)
}

## Generate data for an estimated optimal z basd on true parameters
Gen_post <- function(z_opts, par, num_sim=1e6){
  z1 <- z_opts[[1]]; z2 <- z_opts[[2]]
  
  TB <- TC <- 0
  
  mu_x2 <- par[4]*z2[2]^2 + (par[5] + par[6]*TB + par[7]*TC + par[8]*z2[2])*z2[1] + 
    (par[9] + par[10]*TC + par[11]*z2[3])*(z2[1]^2)
  
  xx1 <- rgamma(num_sim, shape=par[1]*z1[1], rate=par[2])
  mu_x1 <- par[4]*z1[2]^2 + (par[5] + par[6]*TB + par[7]*TC + par[8]*z1[2])*xx1 + 
    (par[9] + par[10]*TC + par[11]*z1[3])*(xx1^2)
  yy1 <- rnorm(num_sim, mean=mu_x1, sd=par[3])
  
  xx2 <- rgamma(num_sim, shape=par[1]*z2[1], rate=par[2])
  mu_x2 <- par[4]*z2[2]^2 + (par[5] + par[6]*TB + par[7]*TC + par[8]*z2[2])*xx2 + 
    (par[9] + par[10]*TC + par[11]*z2[3])*(xx2^2)
  yy2 <- rnorm(num_sim, mean=mu_x2, sd=par[3])
  
  return(list(Y1=yy1, Y2=yy2))  
}


CVR <- function(alp, lb, ub){
  res <- 1/sqrt(1 + 6*(1/alp)*(lb+ub)/((ub-lb)^2))
  
  return(res)
}

SFN1 <- function(alp, lb, ub){
  res <- 1- 1/sqrt(1 + 6*(1/alp)*(lb+ub)/((ub-lb)^2))
  
  return(res)
}

SFN2 <- function(alp, lb, ub){
  res <- 6*(lb+ub)/(alp*(ub-lb)^2 + 6*(lb+ub))
  
  return(res)
}
######### End of functions
