## First cleaning out the memory
rm(list=ls())

## Loading packages, and installs first if package not present
pckgs <- c("latex2exp", "foreach", "doParallel")
absent_pckgs <- pckgs[!(pckgs %in% installed.packages()[,"Package"])]
if(length(absent_pckgs)) install.packages(absent_pckgs, dependencies=TRUE)

lapply(pckgs, library, character.only=TRUE)

################################### Simulation settings ######################################
## Initializing
set.seed(2015)

## Choose internal-noise type:
## 'hom' (for homoscedastic) or 'het' (for heteroscedastic)
noise_type <- 'hom'

## Choose simulation type:
## 'single_tau' or 'multi_tau' corresponding simulation at a single or multiple noise level(s).
sim_type <- 'single_tau'

## noise_level, determines the internal-noise level for 'single_tau' simulations:
## 'low', 'med', or 'high' corresponding to internal noise values of 0, 0.5 and 0.99 respectively.
noise_level <- 'low'

## nRep determines number of repetitions to average over for 'multi_tau' simulations:
## We used 1 or 1e3 in paper. 
## Note that even with parallelization, nRep=1e3 can take a long time to run (~ 4hrs).
nRep <- 1  


## Parameter settings
true_b0 <- 200; true_b1 <- 1; true_b2 <- -1
true_sig <- 0.05; true_gam <- 10

## The domain of Z
LB <- 0; UB <- 1/true_gam

##################################### End simulation settings ######################################

###################################### Begin simulation ############################################
## Applying noise_type setting
if(noise_type=='hom'){
  noisy_input_fname <- "homoskedastic_functions.R"
}else if(noise_type=='het'){
  noisy_input_fname <- "heteroskedastic_functions.R"
}

## Sourcing the right functions
source(noisy_input_fname)

## Running simulation
if(sim_type=='single_tau'){
  ############################# Single_tau simulations ##############################
  
  ## Setting the right noise level based on chosen noise setting
  if(noise_level=='low'){
    inl_setting <- 0
  }else if(noise_level=='med'){
    inl_setting <- 0.5
  }else if(noise_level=='high'){
    inl_setting <- 0.99
  }
  
  ## Finding the tau corresponding to desired internal noise level
  true_tau <- ISFN2(ss=inl_setting, gam=true_gam, lb=LB, ub=UB)
  
  ## Saving the true parameter values
  true_par <- c(true_b0, true_b1, true_b2,
                true_sig, true_gam, true_tau)
  
  ## Data generation
  dd <- Gen_data(par=true_par, num_obs=100, lb=LB, ub=UB)
  pp <- Models(dd)
  
  ## Scale-free noise measure
  sfn1 <- SFN1(true_par, lb=LB, ub=UB)
  sfn2 <- SFN2(true_par, lb=LB, ub=UB)
  
  
  #### Calculating and plotting objective functions
  ## Generating Z's using equi-spaced sequence from domain
  zd <- seq(LB, UB, length.out=1e2)
  
  ## The signal-to-noise ratio
  snr1 <- SNR1(par=pp[[1]], zd)
  
  ## The reference X values for calculating the other objective functions
  xh_ref <- XREF(pars=pp, NN=1e2, lb=LB, ub=UB)
  
  ## The tail probablity objective function
  pdmc1 <- PDMC1(pars=pp, xhRef=xh_ref, zd)
  
  ## The tail expectation objective function
  edmc1 <- EDMC1(pars=pp, xhRef=xh_ref, zd)
  
  
  ## plotting generated data
  par(mar = (c(5, 6, 4, 2) + 0.1), cex.lab=1.5, cex.main=1.5, cex.axis=1.5)
  plot(dd$Z, dd$Y, xlab="True input, Z", ylab="Response, Y",
       main="Marginal Data", col="blue")
  lines(zd, MVK1(par=true_par, zd)$m, col="red", lwd=2)
  lines(zd, MVK1(par=pp[[1]], zd)$m, col="blue", lwd=2, lty=2)
  legend("bottomleft", c("True marginal mean","Estimated marginal mean"),
         lwd=2, col=c("red", "blue"), lty=c(1, 2), cex=1.2)
  
  
  ## Objective function plots
  plot(zd, snr1, type="l", xlab="True input, Z", ylab=TeX("$\\eta^{(1)}(Z)$"),
       main="Estimated signal-to-noise ratio")
  
  plot(zd, pdmc1, type="l", xlab="True input, Z", ylab=TeX("$\\eta^{(2)}(Z)$"),
       main="Estimated tail probability")
  
  plot(zd, edmc1, type="l", xlab="True input, Z", ylab=TeX("$\\eta^{(3)}(Z)$"),
       main="Estimated tail expectation")
  
  par(mar = (c(5, 4, 4, 2) + 0.1))
  
  
  #### Optimal point search
  ## For model 1: accounting for internal noise
  z_1 <- optimize(f=function(x){PDMC1(pars=pp, xhRef=xh_ref, x)},
                  interval=c(0, 0.1), maximum=TRUE)$maximum
  
  ## For model 2: ignoring internal noise
  z_2 <- optimize(f=function(x){PDM2(pars=pp, x)},
                  interval=c(0, 0.1), maximum=TRUE)$maximum
  
  
  #### Evaluation of optimal points: the performance functions
  ## The signal-to-noise ratio
  snrs <- SNR1(par=true_par, c(z_1, z_2))
  
  ## Monte-carlo generation of Y(z_1) and Y(z_2) using true parameters
  YY <- Gen_post(c(z_1, z_2), par=true_par, num_sim=1e5)
  x_star <- -true_b1/(2*true_b2)
  
  ## Y* for evaluating the remaining objective functions
  y_star <- MU.Y(par=true_par, X=x_star)
  
  ## The tail behaviour for MC generated data
  cump1 <- YY[[1]] > y_star
  cump2 <- YY[[2]] > y_star
  
  ## The tail probabilities for MC generated data
  pdms <- c(mean(cump1), mean(cump2))
  
  ## The tail expectations for MC generated data
  edm1 <- ifelse(any(cump1), mean(YY[[1]][cump1]), 0)
  edm2 <- ifelse(any(cump2), mean(YY[[2]][cump2]), 0)
  
  edms <- c(edm1*pdms[1], edm2*pdms[2])
  
  ## The probability of domination for model 1 over model 2, using the MC generated data
  p1p2 <- mean(YY[[1]]>YY[[2]])
  
  ## The expected domination value for model 1 over model 2, using the MC generated data
  e1e2 <- mean(YY[[1]]-YY[[2]])
  
  ## All the quantities
  c(z_1, z_2); snrs; pdms; edms; p1p2; e1e2; sfn1; sfn2
  
}else if(sim_type=='multi_tau'){
  ############################# Multi_tau simulations ##############################
  ## Saving the true parameter values
  true_par <- c(true_b0, true_b1, true_b2,
                true_sig, true_gam, NA)
  
  ## Noise levels
  nn.tau <- 1e2
  s_seq <- seq(0, 1, length.out=nn.tau+1)[-(nn.tau+1)]
  tau.z <- ISFN2(ss=s_seq, gam=true_gam, lb=LB, ub=UB)
  
  #### Setting up parallelization
  ## Calculate the number of logical cores for parallelization
  no_cores <- max(detectCores()-2, 1)
  
  ## Registering the logical cores
  cl<-makeCluster(no_cores)
  registerDoParallel(cl)
  
  ## Running the parallel simulations
  results <- foreach(cc= 1:length(tau.z),
                     .combine=rbind) %dopar% {
                       
                       ## Initialization
                       source(noisy_input_fname)
                       set.seed(2015)
                       
                       ## The current tau value
                       true_par[6] <- tau.z[cc]
                       
                       ## Storage container for simulation results
                       ans <- matrix(NA, nrow=nRep, ncol=10)
                       for(ii in 1:nRep){
                         ## Data generation
                         dd <- Gen_data(par=true_par, num_obs=100, lb=LB, ub=UB)
                         pp <- Models(dd)
                         
                         sfn1a <- SFN1(true_par, lb=LB, ub=UB)
                         sfn1b <- SFN1(pp[[1]], lb=LB, ub=UB)
                         sfn2a <- SFN2(true_par, lb=LB, ub=UB)
                         sfn2b <- SFN2(pp[[1]], lb=LB, ub=UB)
                         
                         
                         #### Optimal point search
                         ## First create reference X-values
                         xh_ref <- XREF(pars=pp, NN=1e2, lb=LB, ub=UB)
                         
                         ## For model 1: accounting for noise
                         z_1 <- optimize(f=function(x){PDMC1(pars=pp, xhRef=xh_ref, x)},
                                         interval=c(0, 0.1), maximum=TRUE)$maximum
                         
                         ## For model 2: ignoring noise
                         z_2 <- optimize(f=function(x){PDM2(pars=pp, x)},
                                         interval=c(0, 0.1), maximum=TRUE)$maximum
                         
                         #### Evaluation of optimal points
                         ## The signal-to-noise ratio
                         snrs <- SNR1(par=true_par, c(z_1, z_2))
                         
                         ## Monte-carlo generation of Y(z_1) and Y(z_2) using true parameters
                         YY <- Gen_post(c(z_1, z_2), par=true_par, num_sim=1e5)
                         x_star <- -true_b1/(2*true_b2)
                         y_star <- MU.Y(par=true_par, X=x_star)
                         
                         ## The tail behaviour for MC generated data
                         cump1 <- YY[[1]] > y_star
                         cump2 <- YY[[2]] > y_star
                         
                         ## The tail probabilities for MC generated data
                         pdms <- c(mean(cump1), mean(cump2))
                         
                         ## The tail expectations for MC generated data
                         edm1 <- ifelse(any(cump1), mean(YY[[1]][cump1]), 0)
                         edm2 <- ifelse(any(cump2), mean(YY[[2]][cump2]), 0)
                         
                         edms <- c(edm1*pdms[1], edm2*pdms[2])
                         
                         ## The probaility of domination for model 1 over model 2, 
                         ## using the MC generated data
                         p1p2 <- mean(YY[[1]]>YY[[2]])
                         
                         ## The expected domination value for model 1 over model 2, 
                         ## using the MC generated data
                         e1e2 <- mean(YY[[1]]-YY[[2]])
                         
                         ans[ii,] <- c(sfn1b, sfn2b, snrs, pdms, edms, p1p2, e1e2)
                       }
                       
                       ## Simulation results for one value of tau
                       res <- c(true_par[6], sfn1a, sfn2a, colMeans(ans))
                       
                       return(res)
                     }
  
  stopCluster(cl)
  
  ## Saving results
  results <- as.data.frame(results)
  names(results) <- c('Tau','True_SFN1', 'True_SFN2','Est_SFN1', 'Est_SFN2',
                      'SNR1','SNR2','PDM1','PDM2','EDM1','EDM2','PDOM','EDIFF')
  results <- results[order(results$Tau),]
  rownames(results) <- NULL
  
  #### Plots
  ## Signal-to-noise ratio vs internal-noise measure
  par(mar = (c(5, 6, 4, 2) + 0.1), cex.lab=1.5, cex.main=1.5, cex.axis=1.5)
  plot(results$True_SFN2, results$SNR1, type="l", ylim=range(results[c('SNR1', 'SNR2')]),
       xlab=TeX("$S(\\gamma, \\tau)$"),
       ylab=TeX("$J(Z(\\tau))$"),
       main="Signal-to-noise ratios", col='red', lwd=2)
  lines(results$True_SFN2, results$SNR2, lty=2, col='blue', lwd=2)
  legend("topright", c(TeX("$J(\\tilde{Z}_{True}(\\tau))$"),
                       TeX("$J(\\tilde{Z}_{Mis}(\\tau))$")),
         lty=c(1,2), col=c("red","blue"), lwd=2, cex=1.5)
  
  ## Approximate tail probability vs internal-noise measure
  plot(results$True_SFN2, results$PDM1, type="l", ylim=c(0,0.5),
       xlab=TeX("$S(\\gamma, \\tau)$"),
       ylab=TeX("$J(Z(\\tau))$"),
       main="Approximate tail probabilities", col='red', lwd=2)
  lines(results$True_SFN2, results$PDM2, lty=2, col='blue', lwd=2)
  legend("topright", c(TeX("$J(\\tilde{Z}_{True}(\\tau))$"),
                       TeX("$J(\\tilde{Z}_{Mis}(\\tau))$")),
         lty=c(1,2), col=c("red","blue"), lwd=2, cex=1.5)
  
  ## Approximate tail expectation vs internal-noise measure
  plot(results$True_SFN2, results$EDM1, type="l", ylim=range(results[c('EDM1', 'EDM2')]),
       xlab=TeX("$S(\\gamma, \\tau)$"),
       ylab=TeX("$J(Z(\\tau))$"),
       main="Approximate tail expectations", col='red', lwd=2)
  lines(results$True_SFN2, results$EDM2, lty=2,  col='blue', lwd=2)
  legend("topright", c(TeX("$J(\\tilde{Z}_{True}(\\tau))$"),
                       TeX("$J(\\tilde{Z}_{Mis}(\\tau))$")),
         lty=c(1,2), col=c("red","blue"), lwd=2, cex=1.5)
  
  ## Approximate probability of domination vs internal-noise measure
  plot(results$True_SFN2, results$PDOM, xlab=TeX("$S(\\gamma, \\tau)$"),
       ylab=TeX("$K(\\tau)$"), main="Approximate probability of domination",
       type="l", col="blue", lwd=2, cex=1.5)
  
  ## Internal-noise measure vs Tau
  plot(results$Tau, results$True_SFN2, type="l", ylim=range(results[c('True_SFN2', 'Est_SFN2')]),
       xlab=TeX("$\\tau$"), ylab=TeX("$S(\\gamma, \\tau)$"),
       main="Scale-Free Noise Measure 2", col='red', lwd=2)
  lines(results$Tau, results$Est_SFN2, lty=2,  col='blue', lwd=2)
  legend("bottomright", c("True","Estimated"), lty=c(1,2), col=c("red","blue"), lwd=2, cex=1.5)
  
  #par(mar = (c(5, 4, 4, 2) + 0.1))
}








